# __init__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:49 PM

"""All Protobuf defnitions."""

from open_gopro.proto.turbo_transfer_pb import RequestSetTurboActive
from open_gopro.proto.response_generic_pb import ResponseGeneric
from open_gopro.proto.request_get_preset_status_pb import RequestGetPresetStatus
from open_gopro.proto.preset_status_pb import NotifyPresetStatus
from open_gopro.proto.set_camera_control_status_pb import RequestSetCameraControlStatus
